@@ -2,7 +2,6 @@
  *
  * SPDX-License-Identifier: BSD-3-Clause */
 #include "pgm_common.h"
-#include "pgm_alternate.h"
 #include "pgm_alternate_trunc_points.h"
 
 /* a struct to store frequently used values. This avoids unnecessary
@@ -36,7 +35,7 @@ typedef struct {
  * [1, 4]. Values are retrieved from a lookup table using binary search, then
  * the final value is calculated using linear interpolation.
  */
-static NPY_INLINE double
+static PGM_INLINE double
 get_truncation_point(double h)
 {
     if (h <= 1) {
@@ -73,18 +72,18 @@ get_truncation_point(double h)
 /*
  * Compute a^L(x|h), the n'th coefficient for the alternating sum S^L(x|h)
  */
-static NPY_INLINE float
+static PGM_INLINE float
 piecewise_coef(unsigned int n, parameter_t const* pr)
 {
     double a = 2 * n + pr->h;
-    double b = n ? pgm_lgamma(n + pr->h) - pr->lgammah : 0;
+    double b = n ? pgm_lgamma(n + pr->h) - pr->lgammah : 0.;
 
     return expf(pr->hlog2 + b - pgm_lgamma(n + 1) - PGM_LS2PI -
                 1.5 * pr->logx - 0.5 * a * a / pr->x) * (float)a;
 }
 
 // compute: k(x|h)
-static NPY_INLINE float
+static PGM_INLINE float
 bounding_kernel(parameter_t const* pr)
 {
     if (pr->x > pr->t) {
@@ -92,7 +91,7 @@ bounding_kernel(parameter_t const* pr)
         return expf(pr->h * a + (pr->h - 1.) * pr->logx -
                     PGM_PI2_8 * pr->x - pr->lgammah);
     }
-    else if (pr->x > 0) {
+    else if (pr->x > 0.) {
         return expf(pr->hlog2 - pr->half_h2 / pr->x -
                     1.5 * pr->logx - PGM_LS2PI) * (float)pr->h;
     }
@@ -102,7 +101,7 @@ bounding_kernel(parameter_t const* pr)
 /*
  * Compute the cdf of the inverse-gaussian distribution.
  */
-static NPY_INLINE float
+static PGM_INLINE float
 invgauss_cdf(parameter_t const* pr)
 {
     static const double sqrt2_inv = 0.7071067811865475f;
@@ -130,14 +129,14 @@ invgauss_cdf(parameter_t const* pr)
  *
  * Notes
  * -----
- * To calculate the probability of sampling on either side of the truncation,
+ * To calculate the probability of sampling on either side of the truncation
  * point we note that:
  * - UpperIncompleteGamma(0.5, x) == sqrt(pi) * erfc(sqrt(x)), the regularized
  *   version of the function, can be written as erfc(sqrt(x)) since the
  *   denominator of the regularized version cancels with the sqrt(pi).
- *   This simplifies the calculation of `p` in the ratio = p / (p + q).
+ *   This simplifies the calculation of `p` when computing p / (p + q).
  */
-static NPY_INLINE void
+static PGM_INLINE void
 set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
 {
     float p, q;
@@ -149,7 +148,7 @@ set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
     pr->lgammah = pgm_lgamma(h);
     pr->hlog2 = h * PGM_LOG2;
 
-    if (!update_params && pr->z > 0) {
+    if (!update_params && pr->z > 0.) {
         pr->h_z = h / pr->z;
         pr->z2 = pr->z * pr->z;
         pr->h_z2 = pr->h_z * pr->h_z;
@@ -157,7 +156,7 @@ set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
         pr->log_lambda_z = logf(pr->lambda_z);
         p = expf(pr->hlog2 - h * pr->z) * invgauss_cdf(pr);
     }
-    else if (pr->z > 0) {
+    else if (pr->z > 0.) {
         pr->h_z = h / pr->z;
         pr->h_z2 = pr->h_z * pr->h_z;
         p = expf(pr->hlog2 - h * pr->z) * invgauss_cdf(pr);
@@ -171,7 +170,7 @@ set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
         p = expf(pr->hlog2) * pgm_erfc(h / sqrt(2. * pr->t));
     }
     q = expf(h * (PGM_LOGPI_2 - pr->log_lambda_z)) *
-        pgm_gammaq(h, pr->lambda_z * pr->t, true);
+        upper_incomplete_gamma(h, pr->lambda_z * pr->t, true);
 
     pr->proposal_probability = q / (p + q);
 }
@@ -187,8 +186,8 @@ set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
  * When mu > t, we use a Scaled-Inverse-Chi-square distribution as a proposal,
  * as explained in [1], page 134. This is equivalent to an Inverse-Gamma with
  * shape=0.5 and scale=lambda/2. We accept the sample only if we sample a
- * uniform less than the acceptance porbability. The probability is
- * exp(-0.5 * z^2 * z). (Refer to Appendix 1 of [1] for a derivation of this probablity).
+ * uniform less than the acceptance probability. The probability is
+ * exp(-0.5 * z^2 * x). (Refer to Appendix 1 of [1] for its derivation).
  *
  * References
  * ----------
@@ -198,7 +197,7 @@ set_sampling_parameters(parameter_t* const pr, double h, bool update_params)
  *      logistic likelihoods.(PhD thesis). Retrieved from
  *      http://hdl.handle.net/2152/21842
  */
-static NPY_INLINE void
+static PGM_INLINE void
 random_right_bounded_invgauss(bitgen_t* bitgen_state, parameter_t* const pr)
 {
     if (pr->t < pr->h_z) {
@@ -227,7 +226,7 @@ random_right_bounded_invgauss(bitgen_t* bitgen_state, parameter_t* const pr)
  * truncated at 1/t (i.e X > 1/t). Then 1/X < t is an Inverse-
  * Gamma right truncated at t. Which is what we want.
  */
-static NPY_INLINE double
+static PGM_INLINE double
 random_jacobi_star(bitgen_t* bitgen_state, parameter_t* const pr)
 {
     for (;;) {
@@ -235,7 +234,7 @@ random_jacobi_star(bitgen_t* bitgen_state, parameter_t* const pr)
             pr->x = random_left_bounded_gamma(bitgen_state, pr->h,
                                               pr->lambda_z, pr->t);
         }
-        else if (pr->z > 0) {
+        else if (pr->z > 0.) {
             random_right_bounded_invgauss(bitgen_state, pr);
         }
         else {
@@ -263,37 +262,62 @@ random_jacobi_star(bitgen_t* bitgen_state, parameter_t* const pr)
     }
 }
 
+void*
+memset(void* __s, int __c, size_t __n);
 /*
  * Sample from PG(h, z) using the alternate method, for h >= 1.
  *
- * For values of h >= 4, we sample J*(h, z/2) = sum(J*(b_i, z/2)) samples such
- * that sum(b_i) = h. Then use the relation PG(h, z) = J*(h, z/2) / 4, to get a
- * sample from the Polya-Gamma distribution.
+ *  Parameters
+ *  ----------
+ *  h : double
+ *      The shape parameter of the distribution. The value must be a positive.
+ *  z : double
+ *      The exponential tilting parameter of the distributon.
+ *  n : size_t
+ *      The number of samples to generate.
+ *  out: array of type double
+ *      The array to place the generated samples. Only the first `n` elements
+ *      will be populated.
  *
- * See: Section 4.3 of Windle et al. (2014)
+ *  Notes
+ *  -----
+ *  For values of h >= 4, we sample J*(h, z/2) = sum(J*(b_i, z/2)) samples such
+ *  that sum(b_i) = h. Then use the relation PG(h, z) = J*(h, z/2) / 4, to get a
+ *  sample from the Polya-Gamma distribution.
+ *
+ *  See: Section 4.3 of Windle et al. (2014)
+ *
+ *  We pre-calculate all values dependant on h only once and avoid
+ *  unnecessary recalculation as long as h remains larger than 4.
  *
- * We pre-calculate all values dependant on h only once and avoid
- * unnecessary recalculation as long as h remains larger than 4.
  */
-double
-random_polyagamma_alternate(bitgen_t *bitgen_state, double h, double z)
+void
+random_polyagamma_alternate(bitgen_t* bitgen_state, double h, double z,
+                            size_t n, double* out)
 {
-    parameter_t pr = {.z = z};
+    parameter_t pr = {.z = 0.5 * fabs(z)};
+    memset(out, 0, n * sizeof(*out));
 
     if (h > pgm_maxh) {
-        double out = 0;
         size_t chunk = h >= (pgm_maxh + 1) ? pgm_maxh : pgm_maxh - 1;
-
         set_sampling_parameters(&pr, chunk, false);
+
         while (h > pgm_maxh) {
-            out += random_jacobi_star(bitgen_state, &pr);
+            for (size_t i = 0; i < n; ++i) {
+                out[i] += 0.25 * random_jacobi_star(bitgen_state, &pr);
+            }
             h -= chunk;
         }
 
         set_sampling_parameters(&pr, h, true);
-        out += random_jacobi_star(bitgen_state, &pr);
-        return 0.25 * out;
+        for (size_t i = 0; i < n; ++i) {
+            out[i] += 0.25 * random_jacobi_star(bitgen_state, &pr);
+        }
+        return;
     }
+
     set_sampling_parameters(&pr, h, false);
-    return 0.25 * random_jacobi_star(bitgen_state, &pr);
+    while (n--) {
+        out[n] += 0.25 * random_jacobi_star(bitgen_state, &pr);
+    }
 }
